#!/usr/bin/env python3
"""
Calendar MCP Server - An MCP server for interacting with macOS Calendar.app

This server provides an interface to Calendar.app on macOS, allowing you to:
- List available calendars
- View, create, update, and delete events
- Search for events by date range
- Get detailed information about calendars and events

The server uses FastMCP to expose both resources and tools that can be used
by AI assistants to interact with the user's calendar.
"""
import datetime
import json
import os
from typing import Dict, List, Optional, Union, Any
import dateparser
from datetime import datetime, timedelta

from mcp.server.fastmcp import FastMCP

from .calendar_store import CalendarStore, CalendarStoreError
from .models import ApiResponse, CalendarEvent, EventCreate, EventUpdate, EventList, CalendarList
from .date_utils import create_date_range, format_iso


# Get port from environment or use default
port = int(os.environ.get("SERVER_PORT", "27212"))

# Create the MCP server with port setting
mcp = FastMCP("Calendar MCP", port=port)


# Resources -----------------------------------------------------------------

@mcp.resource("calendars://list")
def list_calendars() -> str:
    """
    List all available calendars in Calendar.app
    
    Returns:
        JSON string containing calendar names
    """
    try:
        store = CalendarStore(quiet=True)
        calendars = store.get_all_calendars()
        return json.dumps(calendars, ensure_ascii=False)
    except CalendarStoreError as e:
        return json.dumps({"error": str(e)}, ensure_ascii=False)
    except Exception as e:
        return json.dumps({"error": f"Unexpected error: {str(e)}"}, ensure_ascii=False)


@mcp.resource("calendar://{name}")
def get_calendar_info(name: str) -> str:
    """
    Get information about a specific calendar
    
    Args:
        name: The name of the calendar
        
    Returns:
        JSON string with calendar information
    """
    # For now, we just return basic information
    try:
        store = CalendarStore(quiet=True)
        calendars = store.get_all_calendars()
        
        if name not in calendars:
            return json.dumps({"error": f"Calendar '{name}' not found"}, ensure_ascii=False)
        
        return json.dumps({
            "name": name,
            "exists": True
        }, ensure_ascii=False)
    except CalendarStoreError as e:
        return json.dumps({"error": str(e)}, ensure_ascii=False)
    except Exception as e:
        return json.dumps({"error": f"Unexpected error: {str(e)}"}, ensure_ascii=False)


@mcp.resource("events://{calendar_name}")
def get_calendar_events(calendar_name: str) -> str:
    """
    Get events from a specific calendar
    
    Args:
        calendar_name: The name of the calendar
        
    Returns:
        JSON string containing events
    """
    try:
        store = CalendarStore(quiet=True)
        events = store.get_events(calendar_name=calendar_name)
        return json.dumps(events, ensure_ascii=False)
    except CalendarStoreError as e:
        return json.dumps({"error": str(e)}, ensure_ascii=False)
    except Exception as e:
        return json.dumps({"error": f"Unexpected error: {str(e)}"}, ensure_ascii=False)


@mcp.resource("events://{calendar_name}/{start_date}/{end_date}")
def get_calendar_events_by_date_range(
    calendar_name: str, 
    start_date: str, 
    end_date: str
) -> str:
    """
    Get events from a specific calendar within a date range
    
    Args:
        calendar_name: The name of the calendar
        start_date: Start date in format "yyyy-MM-dd"
        end_date: End date in format "yyyy-MM-dd"
        
    Returns:
        JSON string containing events
    """
    try:
        store = CalendarStore(quiet=True)
        events = store.get_events(
            calendar_name=calendar_name,
            start_date=start_date,
            end_date=end_date
        )
        return json.dumps(events, ensure_ascii=False)
    except CalendarStoreError as e:
        return json.dumps({"error": str(e)}, ensure_ascii=False)
    except Exception as e:
        return json.dumps({"error": f"Unexpected error: {str(e)}"}, ensure_ascii=False)


@mcp.resource("event://{calendar_name}/{event_id}")
def get_event(calendar_name: str, event_id: str) -> str:
    """
    Get a specific event by ID
    
    Args:
        calendar_name: The name of the calendar
        event_id: The ID of the event
        
    Returns:
        JSON string containing event details
    """
    try:
        # Get all events and filter by ID
        store = CalendarStore(quiet=True)
        events = store.get_events(calendar_name=calendar_name)
        event = next((e for e in events if e["id"] == event_id), None)
        
        if event:
            return json.dumps(event, ensure_ascii=False)
        else:
            return json.dumps({"error": f"Event with ID '{event_id}' not found"}, ensure_ascii=False)
    except CalendarStoreError as e:
        return json.dumps({"error": str(e)}, ensure_ascii=False)
    except Exception as e:
        return json.dumps({"error": f"Unexpected error: {str(e)}"}, ensure_ascii=False)


# Tools --------------------------------------------------------------------

@mcp.tool()
def list_all_calendars() -> str:
    """
    List all available calendars in Calendar.app
    
    Returns:
        JSON string containing calendar names
    """
    try:
        store = CalendarStore(quiet=True)
        calendars = store.get_all_calendars()
        return json.dumps({
            "calendars": calendars,
            "count": len(calendars)
        }, ensure_ascii=False)
    except CalendarStoreError as e:
        return json.dumps({"error": str(e)}, ensure_ascii=False)
    except Exception as e:
        return json.dumps({"error": f"Unexpected error: {str(e)}"}, ensure_ascii=False)


@mcp.tool()
def search_events(
    query: str,
    calendar_name: Optional[str] = None,
    start_date: Optional[str] = None,
    end_date: Optional[str] = None
) -> str:
    """
    Search for events in Calendar.app
    
    Args:
        query: Search query (case-insensitive substring match)
        calendar_name: (Optional) Specific calendar to search in
        start_date: (Optional) Start date in format "yyyy-MM-dd"
        end_date: (Optional) End date in format "yyyy-MM-dd"
        
    Returns:
        JSON string containing matching events
    """
    try:
        store = CalendarStore(quiet=True)
        events = store.get_events(
            calendar_name=calendar_name,
            start_date=start_date,
            end_date=end_date
        )
        
        # Filter events by query (case-insensitive)
        query = query.lower()
        matching_events = [
            event for event in events
            if (
                query in event["summary"].lower() or
                query in (event["description"] or "").lower() or
                query in (event["location"] or "").lower()
            )
        ]
        
        return json.dumps({
            "events": matching_events,
            "count": len(matching_events)
        }, ensure_ascii=False)
    except CalendarStoreError as e:
        return json.dumps({"error": str(e)}, ensure_ascii=False)
    except Exception as e:
        return json.dumps({"error": f"Unexpected error: {str(e)}"}, ensure_ascii=False)


@mcp.tool()
def create_calendar_event(
    calendar_name: str,
    summary: str,
    start_date: str,
    end_date: str,
    location: Optional[str] = None,
    description: Optional[str] = None
) -> str:
    """
    Create a new event in Calendar.app
    
    Args:
        calendar_name: Name of the calendar to create the event in
        summary: Event title
        start_date: Start date in format "yyyy-MM-ddTHH:mm:ss"
        end_date: End date in format "yyyy-MM-ddTHH:mm:ss"
        location: (Optional) Event location
        description: (Optional) Event description
        
    Returns:
        JSON string containing the result
    """
    try:
        store = CalendarStore(quiet=True)
        event_id = store.create_event(
            calendar_name=calendar_name,
            summary=summary,
            start_date=start_date,
            end_date=end_date,
            location=location,
            description=description
        )
        
        return json.dumps({
            "success": True,
            "event_id": event_id
        }, ensure_ascii=False)
    except CalendarStoreError as e:
        return json.dumps({
            "success": False,
            "error": str(e)
        }, ensure_ascii=False)
    except Exception as e:
        return json.dumps({
            "success": False, 
            "error": f"Unexpected error: {str(e)}"
        }, ensure_ascii=False)


@mcp.tool()
def update_calendar_event(
    event_id: str,
    calendar_name: str,
    summary: Optional[str] = None,
    start_date: Optional[str] = None,
    end_date: Optional[str] = None,
    location: Optional[str] = None,
    description: Optional[str] = None
) -> str:
    """
    Update an existing event in Calendar.app
    
    Args:
        event_id: ID of the event to update
        calendar_name: Name of the calendar containing the event
        summary: (Optional) New event title
        start_date: (Optional) New start date in format "yyyy-MM-ddTHH:mm:ss"
        end_date: (Optional) New end date in format "yyyy-MM-ddTHH:mm:ss"
        location: (Optional) New event location
        description: (Optional) New event description
        
    Returns:
        JSON string containing the result
    """
    try:
        store = CalendarStore(quiet=True)
        success = store.update_event(
            event_id=event_id,
            calendar_name=calendar_name,
            summary=summary,
            start_date=start_date,
            end_date=end_date,
            location=location,
            description=description
        )
        
        return json.dumps({
            "success": success
        }, ensure_ascii=False)
    except CalendarStoreError as e:
        return json.dumps({
            "success": False,
            "error": str(e)
        }, ensure_ascii=False)
    except Exception as e:
        return json.dumps({
            "success": False,
            "error": f"Unexpected error: {str(e)}"
        }, ensure_ascii=False)


@mcp.tool()
def delete_calendar_event(event_id: str, calendar_name: str) -> str:
    """
    Delete an event from Calendar.app
    
    Args:
        event_id: ID of the event to delete
        calendar_name: Name of the calendar containing the event
        
    Returns:
        JSON string containing the result
    """
    try:
        store = CalendarStore(quiet=True)
        success = store.delete_event(
            event_id=event_id,
            calendar_name=calendar_name
        )
        
        return json.dumps({
            "success": success
        }, ensure_ascii=False)
    except CalendarStoreError as e:
        return json.dumps({
            "success": False,
            "error": str(e)
        }, ensure_ascii=False)
    except Exception as e:
        return json.dumps({
            "success": False,
            "error": f"Unexpected error: {str(e)}"
        }, ensure_ascii=False)


@mcp.prompt()
def create_event_prompt(
    calendar_name: str,
    summary: str,
    date: Optional[str] = None,
    start_time: Optional[str] = None,
    end_time: Optional[str] = None,
    duration_minutes: Optional[int] = 60,
    location: Optional[str] = None,
    description: Optional[str] = None
) -> str:
    """
    Create a new event with simplified parameters
    
    Args:
        calendar_name: Name of the calendar to create the event in
        summary: Event title
        date: (Optional) Event date in format "yyyy-MM-dd", defaults to today
        start_time: (Optional) Start time in format "HH:mm", defaults to now
        end_time: (Optional) End time in format "HH:mm"
        duration_minutes: (Optional) Duration in minutes, used if end_time not provided
        location: (Optional) Event location
        description: (Optional) Event description
    """
    # Get current date and time if not provided
    now = datetime.datetime.now()
    
    if not date:
        date = now.strftime("%Y-%m-%d")
        
    if not start_time:
        start_time = now.strftime("%H:%M")
    
    # Calculate end time if not provided
    if not end_time:
        start_dt = datetime.datetime.strptime(f"{date}T{start_time}", "%Y-%m-%dT%H:%M")
        end_dt = start_dt + datetime.timedelta(minutes=duration_minutes)
        end_time = end_dt.strftime("%H:%M")
    
    # Format ISO8601 dates
    start_date = f"{date}T{start_time}:00"
    end_date = f"{date}T{end_time}:00"
    
    # Create the event
    try:
        store = CalendarStore(quiet=True)
        event_id = store.create_event(
            calendar_name=calendar_name,
            summary=summary,
            start_date=start_date,
            end_date=end_date,
            location=location,
            description=description
        )
        
        return f"Event '{summary}' created successfully in calendar '{calendar_name}' on {date} from {start_time} to {end_time}."
    except CalendarStoreError as e:
        return f"Failed to create event: {e}"
    except Exception as e:
        return f"Failed to create event: Unexpected error: {e}"


@mcp.prompt()
def search_events_prompt(
    query: str,
    calendar_name: Optional[str] = None,
    start_date: Optional[str] = None,
    end_date: Optional[str] = None
) -> str:
    """
    Search for events with a user-friendly response
    
    Args:
        query: Search query
        calendar_name: (Optional) Specific calendar to search in
        start_date: (Optional) Start date in format "yyyy-MM-dd"
        end_date: (Optional) End date in format "yyyy-MM-dd"
    """
    try:
        store = CalendarStore(quiet=True)
        events = store.get_events(
            calendar_name=calendar_name,
            start_date=start_date,
            end_date=end_date
        )
        
        # Filter events by query (case-insensitive)
        query = query.lower()
        matching_events = [
            event for event in events
            if (
                query in event["summary"].lower() or
                query in (event["description"] or "").lower() or
                query in (event["location"] or "").lower()
            )
        ]
        
        # Format a human-readable response
        if not matching_events:
            return f"No events found matching '{query}'."
            
        response = f"Found {len(matching_events)} events matching '{query}':\n\n"
        
        for event in matching_events:
            date = event["start"].split("T")[0]
            time = event["start"].split("T")[1][:5]  # Extract HH:MM
            response += f"- {date} {time}: {event['summary']} (in calendar '{event['calendar']}')\n"
            
        return response
    except CalendarStoreError as e:
        return f"Failed to search events: {e}"
    except Exception as e:
        return f"Failed to search events: Unexpected error: {e}"


@mcp.tool()
def get_events_date_range(
    calendar_name: str,
    start_date: Optional[str] = None,
    end_date: Optional[str] = None,
    duration: Optional[str] = None
) -> str:
    """
    Get calendar events within a flexible date range with multiple ways to specify the range.
    
    Args:
        calendar_name: Name of the calendar to get events from
        start_date: Starting date in any format parseable by dateparser (defaults to today)
        end_date: Ending date in any format parseable by dateparser
        duration: Duration from start date or before end date, e.g. "3d", "1 week", "1 month"
                 (default is "3d" if neither end_date nor duration is specified)
    
    The function supports three ways to specify a date range:
    1. start_date only: Returns events from start_date to start_date + 3 days
    2. start_date + duration: Returns events from start_date to start_date + duration
    3. end_date + duration: Returns events from end_date - duration to end_date
    
    Returns:
        JSON string containing events in the specified date range
    """
    try:
        # Parse the duration if provided
        days = 3  # Default duration is 3 days
        if duration:
            # Handle different duration formats
            duration = duration.lower().strip()
            if duration.endswith('d') or duration.endswith(' day') or duration.endswith(' days'):
                # Extract the number part of the duration
                num = ''.join(c for c in duration if c.isdigit())
                if num:
                    days = int(num)
            elif duration.endswith('w') or duration.endswith(' week') or duration.endswith(' weeks'):
                num = ''.join(c for c in duration if c.isdigit())
                if num:
                    days = int(num) * 7
            elif duration.endswith('m') or duration.endswith(' month') or duration.endswith(' months'):
                num = ''.join(c for c in duration if c.isdigit())
                if num:
                    days = int(num) * 30  # Approximation for months
        
        # Determine date range based on provided parameters
        if end_date and not start_date:
            # Case: end_date + duration (backwards from end_date)
            end_dt = dateparser.parse(end_date)
            if not end_dt:
                raise ValueError(f"Could not parse end date: {end_date}")
            start_dt = end_dt - timedelta(days=days)
        else:
            # Case: start_date only or start_date + duration
            if not start_date:
                start_dt = datetime.now().replace(hour=0, minute=0, second=0, microsecond=0)
            else:
                start_dt = dateparser.parse(start_date)
                if not start_dt:
                    raise ValueError(f"Could not parse start date: {start_date}")
            
            if end_date:
                end_dt = dateparser.parse(end_date)
                if not end_dt:
                    raise ValueError(f"Could not parse end date: {end_date}")
            else:
                # Use duration if no end_date is provided
                end_dt = start_dt + timedelta(days=days)
        
        # Ensure dates are in correct order
        if end_dt < start_dt:
            raise ValueError("End date cannot be before start date")
        
        # Format dates as ISO strings for the calendar store
        start_iso = format_iso(start_dt)
        end_iso = format_iso(end_dt)
        
        # Get events from the calendar
        store = CalendarStore(quiet=True)
        events = store.get_events(
            calendar_name=calendar_name,
            start_date=start_iso,
            end_date=end_iso
        )
        
        return json.dumps({
            "success": True,
            "events": events,
            "count": len(events),
            "date_range": {
                "start_date": start_iso,
                "end_date": end_iso,
                "days": (end_dt - start_dt).days
            }
        }, ensure_ascii=False)
    except ValueError as e:
        return json.dumps({
            "success": False,
            "error": f"Date error: {str(e)}"
        }, ensure_ascii=False)
    except CalendarStoreError as e:
        return json.dumps({
            "success": False,
            "error": str(e)
        }, ensure_ascii=False)
    except Exception as e:
        return json.dumps({
            "success": False,
            "error": f"Unexpected error: {str(e)}"
        }, ensure_ascii=False)


# JSON API endpoints -------------------------------------------------------

@mcp.resource("api://calendars")
def api_list_calendars() -> str:
    """
    API endpoint to list all available calendars with JSON response
    
    Returns:
        JSON response with calendars
    """
    try:
        store = CalendarStore(quiet=True)
        calendars = store.get_all_calendars()
        
        response = ApiResponse.success(
            data=CalendarList(calendars=calendars, count=len(calendars))
        )
        return json.dumps(response.model_dump(), ensure_ascii=False)
    except CalendarStoreError as e:
        response = ApiResponse.error(message=str(e))
        return json.dumps(response.model_dump(), ensure_ascii=False)
    except Exception as e:
        response = ApiResponse.error(message=f"Unexpected error: {str(e)}")
        return json.dumps(response.model_dump(), ensure_ascii=False)


@mcp.resource("api://events/{calendar_name}")
def api_get_events(calendar_name: str) -> str:
    """
    API endpoint to get events from a calendar with JSON response
    
    Args:
        calendar_name: The name of the calendar
        
    Returns:
        JSON response with events
    """
    try:
        # Use current date range if not specified
        start_dt, end_dt = create_date_range(None, None)
        
        # Format dates as ISO strings for the calendar store
        start_iso = format_iso(start_dt)
        end_iso = format_iso(end_dt)
        
        store = CalendarStore(quiet=True)
        raw_events = store.get_events(
            calendar_name=calendar_name,
            start_date=start_iso,
            end_date=end_iso
        )
        
        # Convert to validated CalendarEvent objects
        events = []
        for raw_event in raw_events:
            # Add calendar name to each event
            raw_event["calendar_name"] = calendar_name
            event = CalendarEvent(**raw_event)
            events.append(event)
        
        response = ApiResponse.success(
            data=EventList(events=events, count=len(events))
        )
        return json.dumps(response.model_dump(), ensure_ascii=False)
    except ValueError as e:
        response = ApiResponse.error(message=f"Date error: {str(e)}")
        return json.dumps(response.model_dump(), ensure_ascii=False)
    except CalendarStoreError as e:
        response = ApiResponse.error(message=str(e))
        return json.dumps(response.model_dump(), ensure_ascii=False)
    except Exception as e:
        response = ApiResponse.error(message=f"Unexpected error: {str(e)}")
        return json.dumps(response.model_dump(), ensure_ascii=False)


@mcp.resource("api://events/{calendar_name}/{start_date}/{end_date}")
def api_get_events_with_dates(calendar_name: str, start_date: str, end_date: str) -> str:
    """
    API endpoint to get events from a calendar within a date range with JSON response
    
    Args:
        calendar_name: The name of the calendar
        start_date: Start date in any format parseable by dateparser
        end_date: End date in any format parseable by dateparser
        
    Returns:
        JSON response with events
    """
    try:
        # Parse and validate date range
        start_dt, end_dt = create_date_range(start_date, end_date)
        
        # Format dates as ISO strings for the calendar store
        start_iso = format_iso(start_dt)
        end_iso = format_iso(end_dt)
        
        store = CalendarStore(quiet=True)
        raw_events = store.get_events(
            calendar_name=calendar_name,
            start_date=start_iso,
            end_date=end_iso
        )
        
        # Convert to validated CalendarEvent objects
        events = []
        for raw_event in raw_events:
            # Add calendar name to each event
            raw_event["calendar_name"] = calendar_name
            event = CalendarEvent(**raw_event)
            events.append(event)
        
        response = ApiResponse.success(
            data=EventList(events=events, count=len(events))
        )
        return json.dumps(response.model_dump(), ensure_ascii=False)
    except ValueError as e:
        response = ApiResponse.error(message=f"Date error: {str(e)}")
        return json.dumps(response.model_dump(), ensure_ascii=False)
    except CalendarStoreError as e:
        response = ApiResponse.error(message=str(e))
        return json.dumps(response.model_dump(), ensure_ascii=False)
    except Exception as e:
        response = ApiResponse.error(message=f"Unexpected error: {str(e)}")
        return json.dumps(response.model_dump(), ensure_ascii=False)


@mcp.resource("api://events/create/{calendar_name}/{summary}/{start_date}/{end_date}")
def api_create_event_path(calendar_name: str, summary: str, start_date: str, end_date: str) -> str:
    """
    API endpoint to create a new event with JSON response using path parameters
    
    Args:
        calendar_name: Name of the calendar
        summary: Event title
        start_date: Start date in any format parseable by dateparser
        end_date: End date in any format parseable by dateparser
        
    Returns:
        JSON response with result
    """
    try:
        # Validate input with Pydantic model
        event_data = EventCreate(
            calendar_name=calendar_name,
            summary=summary,
            start_date=start_date,
            end_date=end_date,
            location=None,
            description=None,
            all_day=False
        )
        
        store = CalendarStore(quiet=True)
        event_id = store.create_event(
            calendar_name=event_data.calendar_name,
            summary=event_data.summary,
            start_date=format_iso(event_data.start_date),
            end_date=format_iso(event_data.end_date),
            location=None,
            description=None
        )
        
        response = ApiResponse.success(
            data={"event_id": event_id},
            message="Event created successfully"
        )
        return json.dumps(response.model_dump(), ensure_ascii=False)
    except ValueError as e:
        response = ApiResponse.error(message=f"Validation error: {str(e)}")
        return json.dumps(response.model_dump(), ensure_ascii=False)
    except CalendarStoreError as e:
        response = ApiResponse.error(message=str(e))
        return json.dumps(response.model_dump(), ensure_ascii=False)
    except Exception as e:
        response = ApiResponse.error(message=f"Unexpected error: {str(e)}")
        return json.dumps(response.model_dump(), ensure_ascii=False)


@mcp.resource("api://events/update/{event_id}/{calendar_name}")
def api_update_event_path(event_id: str, calendar_name: str) -> str:
    """
    API endpoint to update an event with JSON response using path parameters
    
    Args:
        event_id: ID of the event to update
        calendar_name: Name of the calendar
        
    Returns:
        JSON response with result
    """
    try:
        # Use only the path parameters (no optional updates)
        event_data = EventUpdate(
            event_id=event_id,
            calendar_name=calendar_name
        )
        
        store = CalendarStore(quiet=True)
        
        success = store.update_event(
            event_id=event_data.event_id,
            calendar_name=event_data.calendar_name,
            summary=None,
            start_date=None,
            end_date=None,
            location=None,
            description=None
        )
        
        if success:
            response = ApiResponse.success(message="Event updated successfully")
        else:
            response = ApiResponse.error(message="Failed to update event")
        
        return json.dumps(response.model_dump(), ensure_ascii=False)
    except ValueError as e:
        response = ApiResponse.error(message=f"Validation error: {str(e)}")
        return json.dumps(response.model_dump(), ensure_ascii=False)
    except CalendarStoreError as e:
        response = ApiResponse.error(message=str(e))
        return json.dumps(response.model_dump(), ensure_ascii=False)
    except Exception as e:
        response = ApiResponse.error(message=f"Unexpected error: {str(e)}")
        return json.dumps(response.model_dump(), ensure_ascii=False)


@mcp.resource("api://events/delete/{event_id}/{calendar_name}")
def api_delete_event_path(event_id: str, calendar_name: str) -> str:
    """
    API endpoint to delete an event with JSON response using path parameters
    
    Args:
        event_id: ID of the event to delete
        calendar_name: Name of the calendar
        
    Returns:
        JSON response with result
    """
    try:
        store = CalendarStore(quiet=True)
        success = store.delete_event(event_id=event_id, calendar_name=calendar_name)
        
        if success:
            response = ApiResponse.success(message="Event deleted successfully")
        else:
            response = ApiResponse.error(message="Failed to delete event")
        
        return json.dumps(response.model_dump(), ensure_ascii=False)
    except CalendarStoreError as e:
        response = ApiResponse.error(message=str(e))
        return json.dumps(response.model_dump(), ensure_ascii=False)
    except Exception as e:
        response = ApiResponse.error(message=f"Unexpected error: {str(e)}")
        return json.dumps(response.model_dump(), ensure_ascii=False)


# Main entry point ---------------------------------------------------------

if __name__ == "__main__":
    mcp.run() 